import cython
import cython.cimports.libav as lib
from cython.cimports.av.error import err_check
from cython.cimports.av.video.format import (
    VideoFormatComponent,
    get_pix_fmt,
    get_video_format,
)
from cython.cimports.av.video.frame import VideoFrame
from cython.cimports.cpython import PyBUF_WRITABLE, PyBuffer_FillInfo
from cython.cimports.cpython.buffer import Py_buffer
from cython.cimports.cpython.pycapsule import (
    PyCapsule_GetPointer,
    PyCapsule_IsValid,
    PyCapsule_New,
)
from cython.cimports.libc.stdlib import free, malloc


@cython.cclass
class VideoPlane(Plane):
    def __cinit__(self, frame: VideoFrame, index: cython.int):
        # The palette plane has no associated component or linesize; set fields manually
        fmt = frame.format
        if frame.ptr.hw_frames_ctx:
            frames_ctx: cython.pointer[lib.AVHWFramesContext] = cython.cast(
                cython.pointer[lib.AVHWFramesContext], frame.ptr.hw_frames_ctx.data
            )
            fmt = get_video_format(
                frames_ctx.sw_format, frame.ptr.width, frame.ptr.height
            )

        if index == 1 and fmt.name == "pal8":
            self.width = 256
            self.height = 1
            self.buffer_size = 256 * 4
            return

        for i in range(fmt.ptr.nb_components):
            if fmt.ptr.comp[i].plane == index:
                component = VideoFormatComponent(fmt, i)
                self.width = component.width
                self.height = component.height
                break
        else:
            raise RuntimeError(f"could not find plane {index} of {fmt!r}")

        # Sometimes, linesize is negative (and that is meaningful). We are only
        # insisting that the buffer size be based on the extent of linesize, and
        # ignore it's direction.
        self.buffer_size = abs(self.frame.ptr.linesize[self.index]) * self.height

    @cython.cfunc
    def _buffer_size(self) -> cython.size_t:
        return self.buffer_size

    @property
    def line_size(self):
        """
        Bytes per horizontal line in this plane.

        :type: int
        """
        return self.frame.ptr.linesize[self.index]

    @cython.cfunc
    def _buffer_writable(self) -> cython.bint:
        if self.frame.ptr.hw_frames_ctx:
            return False
        return True

    def __getbuffer__(self, view: cython.pointer[Py_buffer], flags: cython.int):
        if self.frame.ptr.hw_frames_ctx:
            raise TypeError(
                "Hardware frame planes do not support the Python buffer protocol. "
                "Use DLPack (__dlpack__) or download to a software frame."
            )
        if flags & PyBUF_WRITABLE and not self._buffer_writable():
            raise ValueError("buffer is not writable")
        PyBuffer_FillInfo(view, self, self._buffer_ptr(), self._buffer_size(), 0, flags)

    def __dlpack_device__(self):
        if self.frame.ptr.hw_frames_ctx:
            if cython.cast(lib.AVPixelFormat, self.frame.ptr.format) != get_pix_fmt(
                b"cuda"
            ):
                raise NotImplementedError(
                    "DLPack export is only implemented for CUDA hw frames"
                )
            return (kCuda, self.frame._device_id)
        return (kCPU, 0)

    def __dlpack__(self, *, stream: int | None = None):
        if self.frame.ptr.buf[0] == cython.NULL:
            raise TypeError(
                "DLPack export requires a refcounted AVFrame (frame.buf[0] is NULL)"
            )

        sw_fmt: lib.AVPixelFormat
        device_type: cython.int
        device_id: cython.int

        if self.frame.ptr.hw_frames_ctx:
            if cython.cast(lib.AVPixelFormat, self.frame.ptr.format) != get_pix_fmt(
                b"cuda"
            ):
                raise NotImplementedError(
                    "DLPack export is only implemented for CUDA hw frames"
                )
            if stream is not None:
                raise NotImplementedError(
                    "CUDA stream synchronization is not supported. "
                    "Pass stream=None and synchronize before calling __dlpack__."
                )

            frames_ctx: cython.pointer[lib.AVHWFramesContext] = cython.cast(
                cython.pointer[lib.AVHWFramesContext], self.frame.ptr.hw_frames_ctx.data
            )
            sw_fmt = frames_ctx.sw_format
            device_type = kCuda
            device_id = self.frame._device_id
        else:
            sw_fmt = cython.cast(lib.AVPixelFormat, self.frame.ptr.format)
            device_type = kCPU
            device_id = 0

        line_size = self.line_size
        if line_size < 0:
            raise NotImplementedError(
                "negative linesize is not supported for DLPack export"
            )

        nv12 = get_pix_fmt(b"nv12")
        p010le = get_pix_fmt(b"p010le")
        p016le = get_pix_fmt(b"p016le")

        ndim, bits, itemsize = cython.declare(cython.int)
        s0, s1, s2 = cython.declare(int64_t)
        st0, st1, st2 = cython.declare(int64_t)

        if sw_fmt == nv12:
            itemsize = 1
            bits = 8
            if self.index == 0:
                ndim = 2
                s0 = self.frame.ptr.height
                s1 = self.frame.ptr.width
                st0 = line_size
                st1 = 1
            elif self.index == 1:
                ndim = 3
                s0 = self.frame.ptr.height // 2
                s1 = self.frame.ptr.width // 2
                s2 = 2
                st0 = line_size
                st1 = 2
                st2 = 1
            else:
                raise ValueError("invalid plane index for NV12")
        elif sw_fmt == p010le or sw_fmt == p016le:
            itemsize = 2
            bits = 16
            if line_size % itemsize:
                raise ValueError("linesize is not aligned to dtype")
            if self.index == 0:
                ndim = 2
                s0 = self.frame.ptr.height
                s1 = self.frame.ptr.width
                st0 = line_size // itemsize
                st1 = 1
            elif self.index == 1:
                ndim = 3
                s0 = self.frame.ptr.height // 2
                s1 = self.frame.ptr.width // 2
                s2 = 2
                st0 = line_size // itemsize
                st1 = 2
                st2 = 1
            else:
                raise ValueError("invalid plane index for P010/P016")
        else:
            raise NotImplementedError("unsupported sw_format for DLPack export")

        frame_ref: cython.pointer[lib.AVFrame] = lib.av_frame_alloc()
        if frame_ref == cython.NULL:
            raise MemoryError("av_frame_alloc() failed")
        err_check(lib.av_frame_ref(frame_ref, self.frame.ptr))

        shape = cython.cast(
            cython.pointer[int64_t], malloc(ndim * cython.sizeof(int64_t))
        )
        strides = cython.cast(
            cython.pointer[int64_t], malloc(ndim * cython.sizeof(int64_t))
        )
        if shape == cython.NULL or strides == cython.NULL:
            if shape != cython.NULL:
                free(shape)
            if strides != cython.NULL:
                free(strides)
            lib.av_frame_free(cython.address(frame_ref))
            raise MemoryError("malloc() failed")

        if ndim == 2:
            shape[0] = s0
            shape[1] = s1
            strides[0] = st0
            strides[1] = st1
        else:
            shape[0] = s0
            shape[1] = s1
            shape[2] = s2
            strides[0] = st0
            strides[1] = st1
            strides[2] = st2

        ctx = cython.cast(
            cython.pointer[cython.p_void], malloc(3 * cython.sizeof(cython.p_void))
        )
        if ctx == cython.NULL:
            free(shape)
            free(strides)
            lib.av_frame_free(cython.address(frame_ref))
            raise MemoryError("malloc() failed")

        ctx[0] = cython.cast(cython.p_void, frame_ref)
        ctx[1] = cython.cast(cython.p_void, shape)
        ctx[2] = cython.cast(cython.p_void, strides)

        managed = cython.cast(
            cython.pointer[DLManagedTensor], malloc(cython.sizeof(DLManagedTensor))
        )
        if managed == cython.NULL:
            free(ctx)
            free(shape)
            free(strides)
            lib.av_frame_free(cython.address(frame_ref))
            raise MemoryError("malloc() failed")

        managed.dl_tensor.data = cython.cast(cython.p_void, frame_ref.data[self.index])
        managed.dl_tensor.device_type = device_type
        managed.dl_tensor.device_id = device_id
        managed.dl_tensor.ndim = ndim
        managed.dl_tensor.dtype = DLDataType(code=1, bits=bits, lanes=1)
        managed.dl_tensor.shape = shape
        managed.dl_tensor.strides = strides
        managed.dl_tensor.byte_offset = 0
        managed.manager_ctx = cython.cast(cython.p_void, ctx)
        managed.deleter = _dlpack_managed_tensor_deleter

        try:
            capsule = PyCapsule_New(
                cython.cast(cython.p_void, managed),
                b"dltensor",
                _dlpack_capsule_destructor,
            )
        except Exception:
            _dlpack_managed_tensor_deleter(managed)
            raise

        return capsule


@cython.cfunc
@cython.nogil
@cython.exceptval(check=False)
def _dlpack_managed_tensor_deleter(
    managed: cython.pointer[DLManagedTensor],
) -> cython.void:
    if managed == cython.NULL:
        return
    ctx = cython.cast(cython.pointer[cython.p_void], managed.manager_ctx)
    if ctx != cython.NULL:
        frame_ref = cython.cast(cython.pointer[lib.AVFrame], ctx[0])
        shape = cython.cast(cython.pointer[int64_t], ctx[1])
        strides = cython.cast(cython.pointer[int64_t], ctx[2])

        if frame_ref != cython.NULL:
            lib.av_frame_free(cython.address(frame_ref))
        if shape != cython.NULL:
            free(shape)
        if strides != cython.NULL:
            free(strides)
        free(ctx)

    free(managed)


@cython.cfunc
@cython.exceptval(check=False)
def _dlpack_capsule_destructor(capsule: object) -> cython.void:
    if PyCapsule_IsValid(capsule, b"dltensor"):
        managed = cython.cast(
            cython.pointer[DLManagedTensor],
            PyCapsule_GetPointer(capsule, b"dltensor"),
        )
        if managed != cython.NULL:
            managed.deleter(managed)



import weakref
from enum import IntEnum

import cython
import cython.cimports.libav as lib
from cython.cimports.av.codec.codec import Codec
from cython.cimports.av.dictionary import Dictionary
from cython.cimports.av.error import err_check
from cython.cimports.av.video.format import get_video_format


class HWDeviceType(IntEnum):
    none = lib.AV_HWDEVICE_TYPE_NONE
    vdpau = lib.AV_HWDEVICE_TYPE_VDPAU
    cuda = lib.AV_HWDEVICE_TYPE_CUDA
    vaapi = lib.AV_HWDEVICE_TYPE_VAAPI
    dxva2 = lib.AV_HWDEVICE_TYPE_DXVA2
    qsv = lib.AV_HWDEVICE_TYPE_QSV
    videotoolbox = lib.AV_HWDEVICE_TYPE_VIDEOTOOLBOX
    d3d11va = lib.AV_HWDEVICE_TYPE_D3D11VA
    drm = lib.AV_HWDEVICE_TYPE_DRM
    opencl = lib.AV_HWDEVICE_TYPE_OPENCL
    mediacodec = lib.AV_HWDEVICE_TYPE_MEDIACODEC
    vulkan = lib.AV_HWDEVICE_TYPE_VULKAN
    d3d12va = lib.AV_HWDEVICE_TYPE_D3D12VA
    amf = 13  # FFmpeg >=8
    ohcodec = 14
    # TODO: When ffmpeg major is changed, check this enum.


class HWConfigMethod(IntEnum):
    none = 0
    hw_device_ctx = (
        lib.AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX
    )  # This is the only one we support.
    hw_frame_ctx = lib.AV_CODEC_HW_CONFIG_METHOD_HW_FRAMES_CTX
    internal = lib.AV_CODEC_HW_CONFIG_METHOD_INTERNAL
    ad_hoc = lib.AV_CODEC_HW_CONFIG_METHOD_AD_HOC


_cinit_sentinel = cython.declare(object, object())
_singletons = cython.declare(object, weakref.WeakValueDictionary())


@cython.cfunc
def wrap_hwconfig(ptr: cython.pointer[cython.const[lib.AVCodecHWConfig]]) -> HWConfig:
    try:
        return _singletons[cython.cast(cython.Py_ssize_t, ptr)]
    except KeyError:
        pass
    config: HWConfig = HWConfig(_cinit_sentinel)
    config._init(ptr)
    _singletons[cython.cast(cython.Py_ssize_t, ptr)] = config
    return config


@cython.cclass
class HWConfig:
    def __init__(self, sentinel):
        if sentinel is not _cinit_sentinel:
            raise RuntimeError("Cannot instantiate CodecContext")

    @cython.cfunc
    def _init(
        self, ptr: cython.pointer[cython.const[lib.AVCodecHWConfig]]
    ) -> cython.void:
        self.ptr = ptr

    def __repr__(self):
        return (
            f"<av.{self.__class__.__name__} "
            f"device_type={lib.av_hwdevice_get_type_name(self.device_type)} "
            f"format={self.format.name if self.format else None} "
            f"is_supported={self.is_supported} at 0x{cython.cast(cython.Py_ssize_t, self.ptr):x}>"
        )

    @property
    def device_type(self):
        return HWDeviceType(self.ptr.device_type)

    @property
    def format(self):
        return get_video_format(self.ptr.pix_fmt, 0, 0)

    @property
    def methods(self):
        return HWConfigMethod(self.ptr.methods)

    @property
    def is_supported(self):
        return bool(self.ptr.methods & lib.AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX)


@cython.ccall
def hwdevices_available():
    result: list = []
    x: lib.AVHWDeviceType = lib.AV_HWDEVICE_TYPE_NONE
    while True:
        x = lib.av_hwdevice_iterate_types(x)
        if x == lib.AV_HWDEVICE_TYPE_NONE:
            break
        result.append(lib.av_hwdevice_get_type_name(HWDeviceType(x)))
    return result


@cython.cclass
class HWAccel:
    def __init__(
        self,
        device_type,
        device=None,
        allow_software_fallback=True,
        options=None,
        flags=None,
        is_hw_owned=False,
    ):
        if isinstance(device_type, HWDeviceType):
            self._device_type = device_type
        elif isinstance(device_type, str):
            self._device_type = int(lib.av_hwdevice_find_type_by_name(device_type))
        elif isinstance(device_type, int):
            self._device_type = device_type
        else:
            raise ValueError("Unknown type for device_type")

        self.is_hw_owned = is_hw_owned
        self.device_id = 0
        if self._device_type == HWDeviceType.cuda and device:
            self.device_id = int(device)

        self._device = None if device is None else f"{device}"
        self.allow_software_fallback = allow_software_fallback

        self.options = {} if not options else dict(options)
        if self._device_type == HWDeviceType.cuda and self.is_hw_owned:
            self.options.setdefault("primary_ctx", "1")
        self.flags = 0 if not flags else flags
        self.ptr = cython.NULL
        self.config = None

    def _initialize_hw_context(self, codec: Codec):
        config: HWConfig
        for config in codec.hardware_configs:
            if not (config.ptr.methods & lib.AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX):
                continue
            if self._device_type and config.device_type != self._device_type:
                continue
            break
        else:  # nobreak
            raise NotImplementedError(f"No supported hardware config for {codec}")

        self.config = config
        c_device: cython.p_char = cython.NULL
        if self._device:
            device_bytes = self._device.encode()
            c_device = device_bytes
        c_options: Dictionary = Dictionary(self.options)

        err_check(
            lib.av_hwdevice_ctx_create(
                cython.address(self.ptr),
                config.ptr.device_type,
                c_device,
                c_options.ptr,
                self.flags,
            )
        )

    def create(self, codec: Codec) -> HWAccel:
        """Create a new hardware accelerator context with the given codec"""
        if self.ptr:
            raise RuntimeError("Hardware context already initialized")

        ret = HWAccel(
            device_type=self._device_type,
            device=self._device,
            allow_software_fallback=self.allow_software_fallback,
            options=self.options,
            is_hw_owned=self.is_hw_owned,
        )
        ret._initialize_hw_context(codec)
        return ret

    def __dealloc__(self):
        if self.ptr:
            lib.av_buffer_unref(cython.address(self.ptr))



import sys

import cython

container_format_postfix = cython.declare(
    str, "le" if sys.byteorder == "little" else "be"
)
_cinit_bypass_sentinel = cython.declare(object, object())


@cython.cfunc
def get_audio_format(c_format: lib.AVSampleFormat) -> AudioFormat:
    """Get an AudioFormat without going through a string."""

    if c_format < 0:
        return None

    format: AudioFormat = AudioFormat(_cinit_bypass_sentinel)
    format.sample_fmt = c_format
    return format


@cython.cclass
class AudioFormat:
    """Descriptor of audio formats."""

    def __cinit__(self, name):
        if name is _cinit_bypass_sentinel:
            return

        sample_fmt: lib.AVSampleFormat
        if isinstance(name, AudioFormat):
            sample_fmt = cython.cast(AudioFormat, name).sample_fmt
        else:
            sample_fmt = lib.av_get_sample_fmt(name)

        if sample_fmt < 0:
            raise ValueError(f"Not a sample format: {name!r}")

        self.sample_fmt = sample_fmt

    def __repr__(self):
        return f"<av.AudioFormat {self.name}>"

    @property
    def name(self):
        """Canonical name of the sample format.

        >>> AudioFormat('s16p').name
        's16p'

        """
        return lib.av_get_sample_fmt_name(self.sample_fmt)

    @property
    def bytes(self):
        """Number of bytes per sample.

        >>> AudioFormat('s16p').bytes
        2

        """
        return lib.av_get_bytes_per_sample(self.sample_fmt)

    @property
    def bits(self):
        """Number of bits per sample.

        >>> AudioFormat('s16p').bits
        16

        """
        return lib.av_get_bytes_per_sample(self.sample_fmt) << 3

    @property
    def is_planar(self):
        """Is this a planar format?

        Strictly opposite of :attr:`is_packed`.

        """
        return bool(lib.av_sample_fmt_is_planar(self.sample_fmt))

    @property
    def is_packed(self):
        """Is this a packed format?

        Strictly opposite of :attr:`is_planar`.

        """
        return not lib.av_sample_fmt_is_planar(self.sample_fmt)

    @property
    def planar(self):
        """The planar variant of this format.

        Is itself when planar:

        >>> fmt = AudioFormat('s16p')
        >>> fmt.planar is fmt
        True

        """
        if self.is_planar:
            return self
        return get_audio_format(lib.av_get_planar_sample_fmt(self.sample_fmt))

    @property
    def packed(self):
        """The packed variant of this format.

        Is itself when packed:

        >>> fmt = AudioFormat('s16')
        >>> fmt.packed is fmt
        True

        """
        if self.is_packed:
            return self
        return get_audio_format(lib.av_get_packed_sample_fmt(self.sample_fmt))

    @property
    def container_name(self):
        """The name of a :class:`ContainerFormat` which directly accepts this data.

        :raises ValueError: when planar, since there are no such containers.

        """
        if self.is_planar:
            raise ValueError("no planar container formats")

        if self.sample_fmt == lib.AV_SAMPLE_FMT_U8:
            return "u8"
        elif self.sample_fmt == lib.AV_SAMPLE_FMT_S16:
            return "s16" + container_format_postfix
        elif self.sample_fmt == lib.AV_SAMPLE_FMT_S32:
            return "s32" + container_format_postfix
        elif self.sample_fmt == lib.AV_SAMPLE_FMT_FLT:
            return "f32" + container_format_postfix
        elif self.sample_fmt == lib.AV_SAMPLE_FMT_DBL:
            return "f64" + container_format_postfix

        raise ValueError("unknown layout")



cimport libav as lib


cdef extern from "Python.h" nogil:
    void PyErr_PrintEx(int set_sys_last_vars)
    int Py_IsInitialized()
    void PyErr_Display(object, object, object)

cdef extern from "stdio.h" nogil:
    cdef int vsnprintf(char *output, int n, const char *format, lib.va_list args)

cpdef get_last_error()



from enum import IntEnum
from typing import Any, cast

import numpy as np

class VideoEncParamsType(IntEnum):
    NONE = cast(int, ...)
    VP9 = cast(int, ...)
    H264 = cast(int, ...)
    MPEG2 = cast(int, ...)

class VideoEncParams:
    nb_blocks: int
    blocks_offset: int
    block_size: int
    codec_type: VideoEncParamsType
    qp: int
    delta_qp: int
    def block_params(self, idx: int) -> VideoBlockParams: ...
    def qp_map(self) -> np.ndarray[Any, Any]: ...

class VideoBlockParams:
    src_x: int
    src_y: int
    w: int
    h: int
    delta_qp: int



import cython
from cython.cimports.av.packet import Packet
from cython.cimports.av.stream import Stream


@cython.cclass
class SubtitleStream(Stream):
    def __getattr__(self, name):
        return getattr(self.codec_context, name)

    @cython.ccall
    def decode(self, packet: Packet | None = None):
        """
        Decode a :class:`.Packet` and returns a subtitle object.

        :rtype: list[AssSubtitle] | list[BitmapSubtitle]

        .. seealso:: This is a passthrough to :meth:`.CodecContext.decode`.
        """
        if not packet:
            packet = Packet()

        return self.codec_context.decode(packet)



from enum import IntEnum
from typing import cast

from av.codec.codec import Codec
from av.video.format import VideoFormat

class HWDeviceType(IntEnum):
    none = cast(int, ...)
    vdpau = cast(int, ...)
    cuda = cast(int, ...)
    vaapi = cast(int, ...)
    dxva2 = cast(int, ...)
    qsv = cast(int, ...)
    videotoolbox = cast(int, ...)
    d3d11va = cast(int, ...)
    drm = cast(int, ...)
    opencl = cast(int, ...)
    mediacodec = cast(int, ...)
    vulkan = cast(int, ...)
    d3d12va = cast(int, ...)

class HWConfigMethod(IntEnum):
    none = cast(int, ...)
    hw_device_ctx = cast(int, ...)
    hw_frame_ctx = cast(int, ...)
    internal = cast(int, ...)
    ad_hoc = cast(int, ...)

class HWConfig:
    @property
    def device_type(self) -> HWDeviceType: ...
    @property
    def format(self) -> VideoFormat: ...
    @property
    def methods(self) -> HWConfigMethod: ...
    @property
    def is_supported(self) -> bool: ...

class HWAccel:
    options: dict[str, object]

    @property
    def is_hw_owned(self) -> bool: ...
    @property
    def device_id(self) -> int: ...
    def __init__(
        self,
        device_type: str | HWDeviceType,
        device: str | int | None = None,
        allow_software_fallback: bool = False,
        options: dict[str, object] | None = None,
        flags: int | None = None,
        is_hw_owned: bool = False,
    ) -> None: ...
    def create(self, codec: Codec) -> HWAccel: ...

def hwdevices_available() -> list[str]: ...



import cython
from cython.cimports.av.audio.frame import AudioFrame
from cython.cimports.av.packet import Packet


@cython.cclass
class AudioStream(Stream):
    def __repr__(self):
        if self.codec_context is None:
            return f"<av.AudioStream #{self.index} audio/<nocodec> at 0x{id(self):x}>"
        form = self.format.name if self.format else None
        return (
            f"<av.AudioStream #{self.index} {self.name} at {self.rate}Hz,"
            f" {self.layout.name}, {form} at 0x{id(self):x}>"
        )

    def __getattr__(self, name):
        if self.codec_context is None:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )
        return getattr(self.codec_context, name)

    @cython.ccall
    def encode(self, frame: AudioFrame | None = None):
        """
        Encode an :class:`.AudioFrame` and return a list of :class:`.Packet`.

        :rtype: list[Packet]

        .. seealso:: This is mostly a passthrough to :meth:`.CodecContext.encode`.
        """

        packets = self.codec_context.encode(frame)
        packet: Packet
        for packet in packets:
            packet._stream = self
            packet.ptr.stream_index = self.ptr.index

        return packets

    @cython.ccall
    def decode(self, packet: Packet | None = None):
        """
        Decode a :class:`.Packet` and return a list of :class:`.AudioFrame`.

        :rtype: list[AudioFrame]

        .. seealso:: This is a passthrough to :meth:`.CodecContext.decode`.
        """

        return self.codec_context.decode(packet)



import cython
from cython.cimports.av.audio.format import get_audio_format
from cython.cimports.av.audio.layout import get_audio_layout
from cython.cimports.av.audio.plane import AudioPlane
from cython.cimports.av.error import err_check
from cython.cimports.av.utils import check_ndarray

_cinit_bypass_sentinel = cython.declare(object, object())


@cython.cfunc
def alloc_audio_frame() -> AudioFrame:
    return AudioFrame(_cinit_bypass_sentinel)


format_dtypes = {
    "dbl": "f8",
    "dblp": "f8",
    "flt": "f4",
    "fltp": "f4",
    "s16": "i2",
    "s16p": "i2",
    "s32": "i4",
    "s32p": "i4",
    "u8": "u1",
    "u8p": "u1",
}


@cython.cclass
class AudioFrame(Frame):
    """A frame of audio."""

    def __cinit__(self, format="s16", layout="stereo", samples=0, align=1):
        if format is _cinit_bypass_sentinel:
            return

        cy_format: AudioFormat = AudioFormat(format)
        cy_layout: AudioLayout = AudioLayout(layout)
        self._init(cy_format.sample_fmt, cy_layout.layout, samples, align)

    @cython.cfunc
    def _init(
        self,
        format: lib.AVSampleFormat,
        layout: lib.AVChannelLayout,
        nb_samples: cython.uint,
        align: cython.uint,
    ):
        self.ptr.nb_samples = nb_samples
        self.ptr.format = format
        self.ptr.ch_layout = layout

        # Sometimes this is called twice. Oh well.
        self._init_user_attributes()

        if self.layout.nb_channels != 0 and nb_samples:
            # Cleanup the old buffer.
            lib.av_freep(cython.address(self._buffer))

            # Get a new one.
            self._buffer_size = err_check(
                lib.av_samples_get_buffer_size(
                    cython.NULL, self.layout.nb_channels, nb_samples, format, align
                )
            )
            self._buffer = cython.cast(
                cython.pointer[uint8_t], lib.av_malloc(self._buffer_size)
            )
            if not self._buffer:
                raise MemoryError("cannot allocate AudioFrame buffer")

            # Connect the data pointers to the buffer.
            err_check(
                lib.avcodec_fill_audio_frame(
                    self.ptr,
                    self.layout.nb_channels,
                    cython.cast(lib.AVSampleFormat, self.ptr.format),
                    self._buffer,
                    self._buffer_size,
                    align,
                )
            )

    def __dealloc__(self):
        lib.av_freep(cython.address(self._buffer))

    @cython.cfunc
    def _init_user_attributes(self):
        self.layout = get_audio_layout(self.ptr.ch_layout)
        self.format = get_audio_format(cython.cast(lib.AVSampleFormat, self.ptr.format))

    def __repr__(self):
        return (
            f"<av.{self.__class__.__name__} pts={self.pts}, {self.samples} "
            f"samples at {self.rate}Hz, {self.layout.name}, {self.format.name} at 0x{id(self):x}>"
        )

    @staticmethod
    def from_ndarray(array, format="s16", layout="stereo"):
        """
        Construct a frame from a numpy array.
        """
        import numpy as np

        py_format = format if isinstance(format, AudioFormat) else AudioFormat(format)
        py_layout = layout if isinstance(layout, AudioLayout) else AudioLayout(layout)
        format = py_format.name

        # map avcodec type to numpy type
        try:
            dtype = np.dtype(format_dtypes[format])
        except KeyError:
            raise ValueError(
                f"Conversion from numpy array with format `{format}` is not yet supported"
            )

        # check input format
        nb_channels = py_layout.nb_channels
        check_ndarray(array, dtype, 2)
        if py_format.is_planar:
            if array.shape[0] != nb_channels:
                raise ValueError(
                    f"Expected planar `array.shape[0]` to equal `{nb_channels}` but got `{array.shape[0]}`"
                )
            samples = array.shape[1]
        else:
            if array.shape[0] != 1:
                raise ValueError(
                    f"Expected packed `array.shape[0]` to equal `1` but got `{array.shape[0]}`"
                )
            samples = array.shape[1] // nb_channels

        frame = AudioFrame(format=py_format, layout=py_layout, samples=samples)
        for i, plane in enumerate(frame.planes):
            plane.update(array[i, :])
        return frame

    @property
    def planes(self):
        """
        A tuple of :class:`~av.audio.plane.AudioPlane`.

        :type: tuple
        """
        plane_count: cython.int = 0
        while self.ptr.extended_data[plane_count]:
            plane_count += 1

        return tuple([AudioPlane(self, i) for i in range(plane_count)])

    @property
    def samples(self):
        """
        Number of audio samples (per channel).

        :type: int
        """
        return self.ptr.nb_samples

    @property
    def sample_rate(self):
        """
        Sample rate of the audio data, in samples per second.

        :type: int
        """
        return self.ptr.sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        self.ptr.sample_rate = value

    @property
    def rate(self):
        """Another name for :attr:`sample_rate`."""
        return self.ptr.sample_rate

    @rate.setter
    def rate(self, value):
        self.ptr.sample_rate = value

    def to_ndarray(self):
        """Get a numpy array of this frame.

        .. note:: Numpy must be installed.

        """
        import numpy as np

        try:
            dtype = np.dtype(format_dtypes[self.format.name])
        except KeyError:
            raise ValueError(
                f"Conversion from {self.format.name!r} format to numpy array is not supported."
            )

        if self.format.is_planar:
            count = self.samples
        else:
            count = self.samples * self.layout.nb_channels

        return np.vstack(
            [np.frombuffer(x, dtype=dtype, count=count) for x in self.planes]
        )



import cython
from cython.cimports import libav as lib
from cython.cimports.av.audio.format import AudioFormat, get_audio_format
from cython.cimports.av.audio.frame import AudioFrame, alloc_audio_frame
from cython.cimports.av.audio.layout import AudioLayout, get_audio_layout
from cython.cimports.av.frame import Frame
from cython.cimports.av.packet import Packet


@cython.cclass
class AudioCodecContext(CodecContext):
    @cython.cfunc
    def _prepare_frames_for_encode(self, input_frame: Frame | None) -> list:
        frame: AudioFrame | None = input_frame
        allow_var_frame_size: cython.bint = (
            self.ptr.codec.capabilities & lib.AV_CODEC_CAP_VARIABLE_FRAME_SIZE
        )

        # Note that the resampler will simply return an input frame if there is
        # no resampling to be done. The control flow was just a little easier this way.
        if not self.resampler:
            self.resampler = AudioResampler(
                format=self.format,
                layout=self.layout,
                rate=self.ptr.sample_rate,
                frame_size=None if allow_var_frame_size else self.ptr.frame_size,
            )
        frames = self.resampler.resample(frame)
        if input_frame is None:
            frames.append(None)  # flush if input frame is None

        return frames

    @cython.cfunc
    def _alloc_next_frame(self) -> Frame:
        return alloc_audio_frame()

    @cython.cfunc
    def _setup_decoded_frame(self, frame: Frame, packet: Packet):
        CodecContext._setup_decoded_frame(self, frame, packet)
        aframe: AudioFrame = frame
        aframe._init_user_attributes()

    @property
    def frame_size(self):
        """
        Number of samples per channel in an audio frame.

        :type: int
        """
        return self.ptr.frame_size

    @property
    def sample_rate(self):
        """
        Sample rate of the audio data, in samples per second.

        :type: int
        """
        return self.ptr.sample_rate

    @sample_rate.setter
    def sample_rate(self, value: cython.int):
        self.ptr.sample_rate = value

    @property
    def rate(self):
        """Another name for :attr:`sample_rate`."""
        return self.sample_rate

    @rate.setter
    def rate(self, value):
        self.sample_rate = value

    @property
    def channels(self):
        return self.layout.nb_channels

    @property
    def layout(self):
        """
        The audio channel layout.

        :type: AudioLayout
        """
        return get_audio_layout(self.ptr.ch_layout)

    @layout.setter
    def layout(self, value):
        layout: AudioLayout = AudioLayout(value)
        self.ptr.ch_layout = layout.layout

    @property
    def format(self):
        """
        The audio sample format.

        :type: AudioFormat
        """
        return get_audio_format(self.ptr.sample_fmt)

    @format.setter
    def format(self, value):
        format: AudioFormat = AudioFormat(value)
        self.ptr.sample_fmt = format.sample_fmt



import os
import time
from enum import Flag, IntEnum
from pathlib import Path

import cython
import cython.cimports.libav as lib
from cython.cimports.av.container.core import timeout_info
from cython.cimports.av.container.input import InputContainer
from cython.cimports.av.container.output import OutputContainer
from cython.cimports.av.container.pyio import pyio_close_custom_gil, pyio_close_gil
from cython.cimports.av.error import err_check, stash_exception
from cython.cimports.av.format import build_container_format
from cython.cimports.av.utils import (
    avdict_to_dict,
    avrational_to_fraction,
    dict_to_avdict,
    to_avrational,
)
from cython.cimports.libc.stdint import int64_t
from cython.operator import dereference

from av.logging import Capture as LogCapture

_cinit_sentinel = cython.declare(object, object())

AVChapterPtrPtr = cython.typedef(cython.pointer[cython.pointer[lib.AVChapter]])


@cython.cfunc
@cython.nogil
@cython.exceptval(check=False)
def interrupt_cb(p: cython.p_void) -> cython.int:
    info: timeout_info = dereference(cython.cast(cython.pointer[timeout_info], p))
    if info.timeout < 0:  # timeout < 0 means no timeout
        return 0

    current_time: cython.double
    with cython.gil:
        current_time = time.monotonic()
        if current_time < info.start_time:
            # Raise this when we get back to Python.
            stash_exception(
                (
                    RuntimeError,
                    RuntimeError("Clock has been changed to before timeout start"),
                    None,
                )
            )
            return 1

    if current_time > info.start_time + info.timeout:
        return 1
    return 0


@cython.cfunc
@cython.nogil
@cython.exceptval(check=False)
def pyav_io_open(
    s: cython.pointer[lib.AVFormatContext],
    pb: cython.pointer[cython.pointer[lib.AVIOContext]],
    url: cython.p_const_char,
    flags: cython.int,
    options: cython.pointer[cython.pointer[lib.AVDictionary]],
) -> cython.int:
    with cython.gil:
        return pyav_io_open_gil(s, pb, url, flags, options)


@cython.cfunc
@cython.exceptval(check=False)
def pyav_io_open_gil(
    s: cython.pointer[lib.AVFormatContext],
    pb: cython.pointer[cython.pointer[lib.AVIOContext]],
    url: cython.p_const_char,
    flags: cython.int,
    options: cython.pointer[cython.pointer[lib.AVDictionary]],
) -> cython.int:
    container: Container
    file: object
    pyio_file: PyIOFile
    try:
        container = cython.cast(Container, dereference(s).opaque)

        if options is not cython.NULL:
            options_dict = avdict_to_dict(
                dereference(
                    cython.cast(
                        cython.pointer[cython.pointer[lib.AVDictionary]], options
                    )
                ),
                encoding=container.metadata_encoding,
                errors=container.metadata_errors,
            )
        else:
            options_dict = {}

        file = container.io_open(
            cython.cast(str, url) if url is not cython.NULL else "", flags, options_dict
        )

        pyio_file = PyIOFile(
            file, container.buffer_size, (flags & lib.AVIO_FLAG_WRITE) != 0
        )

        # Add it to the container to avoid it being deallocated
        container.open_files[cython.cast(int64_t, pyio_file.iocontext.opaque)] = (
            pyio_file
        )
        pb[0] = pyio_file.iocontext
        return 0
    except Exception:
        return stash_exception()


@cython.cfunc
@cython.nogil
@cython.exceptval(check=False)
def pyav_io_close(
    s: cython.pointer[lib.AVFormatContext], pb: cython.pointer[lib.AVIOContext]
) -> cython.int:
    with cython.gil:
        return pyav_io_close_gil(s, pb)


@cython.cfunc
@cython.exceptval(check=False)
def pyav_io_close_gil(
    s: cython.pointer[lib.AVFormatContext], pb: cython.pointer[lib.AVIOContext]
) -> cython.int:
    container: Container
    result: cython.int = 0
    try:
        container = cython.cast(Container, dereference(s).opaque)
        if (
            container.open_files is not None
            and cython.cast(int64_t, pb.opaque) in container.open_files
        ):
            result = pyio_close_custom_gil(pb)

            # Remove it from the container so that it can be deallocated
            del container.open_files[cython.cast(int64_t, pb.opaque)]
        else:
            result = pyio_close_gil(pb)

    except Exception:
        stash_exception()
        result = lib.AVERROR_UNKNOWN  # Or another appropriate error code

    return result


@cython.cfunc
@cython.nogil
@cython.exceptval(check=False)
def _free_chapters(ctx: cython.pointer[lib.AVFormatContext]) -> cython.void:
    i: cython.Py_ssize_t
    if ctx.chapters != cython.NULL:
        for i in range(ctx.nb_chapters):
            if ctx.chapters[i] != cython.NULL:
                if ctx.chapters[i].metadata != cython.NULL:
                    lib.av_dict_free(cython.address(ctx.chapters[i].metadata))
                lib.av_freep(
                    cython.cast(cython.pp_void, cython.address(ctx.chapters[i]))
                )
        lib.av_freep(cython.cast(cython.pp_void, cython.address(ctx.chapters)))
    ctx.nb_chapters = 0


# fmt: off
class Flags(Flag):
    gen_pts: "Generate missing pts even if it requires parsing future frames." = lib.AVFMT_FLAG_GENPTS
    ign_idx: "Ignore index." = lib.AVFMT_FLAG_IGNIDX
    non_block: "Do not block when reading packets from input." = lib.AVFMT_FLAG_NONBLOCK
    ign_dts: "Ignore DTS on frames that contain both DTS & PTS." = lib.AVFMT_FLAG_IGNDTS
    no_fillin: "Do not infer any values from other values, just return what is stored in the container." = lib.AVFMT_FLAG_NOFILLIN
    no_parse: "Do not use AVParsers, you also must set AVFMT_FLAG_NOFILLIN as the fill in code works on frames and no parsing -> no frames. Also seeking to frames can not work if parsing to find frame boundaries has been disabled." = lib.AVFMT_FLAG_NOPARSE
    no_buffer: "Do not buffer frames when possible." = lib.AVFMT_FLAG_NOBUFFER
    custom_io: "The caller has supplied a custom AVIOContext, don't avio_close() it." = lib.AVFMT_FLAG_CUSTOM_IO
    discard_corrupt: "Discard frames marked corrupted." = lib.AVFMT_FLAG_DISCARD_CORRUPT
    flush_packets: "Flush the AVIOContext every packet." = lib.AVFMT_FLAG_FLUSH_PACKETS
    bitexact: "When muxing, try to avoid writing any random/volatile data to the output. This includes any random IDs, real-time timestamps/dates, muxer version, etc. This flag is mainly intended for testing." = lib.AVFMT_FLAG_BITEXACT
    sort_dts: "Try to interleave outputted packets by dts (using this flag can slow demuxing down)." = lib.AVFMT_FLAG_SORT_DTS
    fast_seek: "Enable fast, but inaccurate seeks for some formats." = lib.AVFMT_FLAG_FAST_SEEK
    auto_bsf: "Add bitstream filters as requested by the muxer." = lib.AVFMT_FLAG_AUTO_BSF

class AudioCodec(IntEnum):
    """Enumeration for audio codec IDs."""
    none = lib.AV_CODEC_ID_NONE  # No codec.
    pcm_alaw = lib.AV_CODEC_ID_PCM_ALAW  # PCM A-law.
    pcm_bluray = lib.AV_CODEC_ID_PCM_BLURAY  # PCM Blu-ray.
    pcm_dvd = lib.AV_CODEC_ID_PCM_DVD  # PCM DVD.
    pcm_f16le = lib.AV_CODEC_ID_PCM_F16LE  # PCM F16 little-endian.
    pcm_f24le = lib.AV_CODEC_ID_PCM_F24LE  # PCM F24 little-endian.
    pcm_f32be = lib.AV_CODEC_ID_PCM_F32BE  # PCM F32 big-endian.
    pcm_f32le = lib.AV_CODEC_ID_PCM_F32LE  # PCM F32 little-endian.
    pcm_f64be = lib.AV_CODEC_ID_PCM_F64BE  # PCM F64 big-endian.
    pcm_f64le = lib.AV_CODEC_ID_PCM_F64LE  # PCM F64 little-endian.
    pcm_lxf = lib.AV_CODEC_ID_PCM_LXF  # PCM LXF.
    pcm_mulaw = lib.AV_CODEC_ID_PCM_MULAW  # PCM μ-law.
    pcm_s16be = lib.AV_CODEC_ID_PCM_S16BE  # PCM signed 16-bit big-endian.
    pcm_s16be_planar = lib.AV_CODEC_ID_PCM_S16BE_PLANAR  # PCM signed 16-bit big-endian planar.
    pcm_s16le = lib.AV_CODEC_ID_PCM_S16LE  # PCM signed 16-bit little-endian.
    pcm_s16le_planar = lib.AV_CODEC_ID_PCM_S16LE_PLANAR  # PCM signed 16-bit little-endian planar.
    pcm_s24be = lib.AV_CODEC_ID_PCM_S24BE  # PCM signed 24-bit big-endian.
    pcm_s24daud = lib.AV_CODEC_ID_PCM_S24DAUD  # PCM signed 24-bit D-Cinema audio.
    pcm_s24le = lib.AV_CODEC_ID_PCM_S24LE  # PCM signed 24-bit little-endian.
    pcm_s24le_planar = lib.AV_CODEC_ID_PCM_S24LE_PLANAR  # PCM signed 24-bit little-endian planar.
    pcm_s32be = lib.AV_CODEC_ID_PCM_S32BE  # PCM signed 32-bit big-endian.
    pcm_s32le = lib.AV_CODEC_ID_PCM_S32LE  # PCM signed 32-bit little-endian.
    pcm_s32le_planar = lib.AV_CODEC_ID_PCM_S32LE_PLANAR  # PCM signed 32-bit little-endian planar.
    pcm_s64be = lib.AV_CODEC_ID_PCM_S64BE  # PCM signed 64-bit big-endian.
    pcm_s64le = lib.AV_CODEC_ID_PCM_S64LE  # PCM signed 64-bit little-endian.
    pcm_s8 = lib.AV_CODEC_ID_PCM_S8  # PCM signed 8-bit.
    pcm_s8_planar = lib.AV_CODEC_ID_PCM_S8_PLANAR  # PCM signed 8-bit planar.
    pcm_u16be = lib.AV_CODEC_ID_PCM_U16BE  # PCM unsigned 16-bit big-endian.
    pcm_u16le = lib.AV_CODEC_ID_PCM_U16LE  # PCM unsigned 16-bit little-endian.
    pcm_u24be = lib.AV_CODEC_ID_PCM_U24BE  # PCM unsigned 24-bit big-endian.
    pcm_u24le = lib.AV_CODEC_ID_PCM_U24LE  # PCM unsigned 24-bit little-endian.
    pcm_u32be = lib.AV_CODEC_ID_PCM_U32BE  # PCM unsigned 32-bit big-endian.
    pcm_u32le = lib.AV_CODEC_ID_PCM_U32LE  # PCM unsigned 32-bit little-endian.
    pcm_u8 = lib.AV_CODEC_ID_PCM_U8  # PCM unsigned 8-bit.
    pcm_vidc = lib.AV_CODEC_ID_PCM_VIDC  # PCM VIDC.
# fmt: on


@cython.cclass
class Container:
    def __cinit__(
        self,
        sentinel,
        file_,
        format_name,
        options,
        container_options,
        stream_options,
        hwaccel,
        metadata_encoding,
        metadata_errors,
        buffer_size,
        open_timeout,
        read_timeout,
        io_open,
    ):
        if sentinel is not _cinit_sentinel:
            raise RuntimeError("cannot construct base Container")

        writeable: cython.bint = isinstance(self, OutputContainer)
        if not writeable and not isinstance(self, InputContainer):
            raise RuntimeError("Container cannot be directly extended.")

        if isinstance(file_, str):
            self.name = file_
        else:
            self.name = str(getattr(file_, "name", "<none>"))

        self.options = dict(options or ())
        self.container_options = dict(container_options or ())
        self.stream_options = [dict(x) for x in stream_options or ()]

        self.hwaccel = hwaccel

        self.metadata_encoding = metadata_encoding
        self.metadata_errors = metadata_errors

        self.open_timeout = open_timeout
        self.read_timeout = read_timeout

        self.buffer_size = buffer_size
        self.io_open = io_open

        acodec = None  # no audio codec specified
        if format_name is not None:
            if ":" in format_name:
                format_name, acodec = format_name.split(":")
            self.format = ContainerFormat(format_name)

        res: cython.int
        name_obj: bytes = os.fsencode(self.name)
        name: cython.p_char = name_obj
        ofmt: cython.pointer[cython.const[lib.AVOutputFormat]]

        if writeable:
            self._myflag |= 1  # enum.writeable = True
            ofmt = (
                self.format.optr
                if self.format
                else lib.av_guess_format(cython.NULL, name, cython.NULL)
            )
            if ofmt == cython.NULL:
                raise ValueError("Could not determine output format")

            with cython.nogil:
                # This does not actually open the file.
                res = lib.avformat_alloc_output_context2(
                    cython.address(self.ptr),
                    ofmt,
                    cython.NULL,
                    name,
                )
            self.err_check(res)
        else:
            # We need the context before we open the input AND setup Python IO.
            self.ptr = lib.avformat_alloc_context()

            # Setup interrupt callback
            if self.open_timeout is not None or self.read_timeout is not None:
                self.ptr.interrupt_callback.callback = interrupt_cb
                self.ptr.interrupt_callback.opaque = cython.address(
                    self.interrupt_callback_info
                )

            if acodec is not None:
                self.ptr.audio_codec_id = getattr(AudioCodec, acodec)

        self.ptr.flags |= lib.AVFMT_FLAG_GENPTS
        self.ptr.opaque = cython.cast(cython.p_void, self)

        # Setup Python IO.
        self.open_files = {}
        if not isinstance(file_, basestring):
            self.file = PyIOFile(file_, buffer_size, writeable)
            self.ptr.pb = self.file.iocontext

        if io_open is not None:
            self.ptr.io_open = pyav_io_open
            self.ptr.io_close2 = pyav_io_close
            self.ptr.flags |= lib.AVFMT_FLAG_CUSTOM_IO

        ifmt: cython.pointer[cython.const[lib.AVInputFormat]]
        c_options: Dictionary
        if not writeable:
            ifmt = self.format.iptr if self.format else cython.NULL
            c_options = Dictionary(self.options, self.container_options)

            self.set_timeout(self.open_timeout)
            self.start_timeout()
            with cython.nogil:
                res = lib.avformat_open_input(
                    cython.address(self.ptr), name, ifmt, cython.address(c_options.ptr)
                )
            self.set_timeout(None)
            self.err_check(res)
            self._myflag |= 2  # enum.input_was_opened = True

        if format_name is None:
            self.format = build_container_format(self.ptr.iformat, self.ptr.oformat)

    def __dealloc__(self):
        with cython.nogil:
            lib.avformat_free_context(self.ptr)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        raise NotImplementedError

    def __repr__(self):
        return f"<av.{self.__class__.__name__} {self.file or self.name!r}>"

    @cython.cfunc
    @cython.exceptval(-1, check=False)
    def err_check(self, value: cython.int) -> cython.int:
        return err_check(value, filename=self.name)

    def dumps_format(self):
        self._assert_open()
        with LogCapture() as logs:
            lib.av_dump_format(self.ptr, 0, "", isinstance(self, OutputContainer))
        return "".join(log[2] for log in logs)

    @cython.cfunc
    def set_timeout(self, timeout):
        if timeout is None:
            self.interrupt_callback_info.timeout = -1.0
        else:
            self.interrupt_callback_info.timeout = timeout

    @cython.cfunc
    def start_timeout(self):
        self.interrupt_callback_info.start_time = time.monotonic()

    @cython.cfunc
    def _assert_open(self):
        if self.ptr == cython.NULL:
            raise AssertionError("Container is not open")

    @property
    def flags(self):
        self._assert_open()
        return self.ptr.flags

    @flags.setter
    def flags(self, value: cython.int):
        self._assert_open()
        self.ptr.flags = value

    @property
    def input_was_opened(self):
        return self._myflag & 2

    def chapters(self):
        self._assert_open()
        result: list = []
        i: cython.Py_ssize_t
        for i in range(self.ptr.nb_chapters):
            ch = self.ptr.chapters[i]
            result.append(
                {
                    "id": ch.id,
                    "start": ch.start,
                    "end": ch.end,
                    "time_base": avrational_to_fraction(cython.address(ch.time_base)),
                    "metadata": avdict_to_dict(
                        ch.metadata, self.metadata_encoding, self.metadata_errors
                    ),
                }
            )
        return result

    def set_chapters(self, chapters):
        self._assert_open()

        count: cython.Py_ssize_t = len(chapters)
        i: cython.Py_ssize_t
        ch_array: AVChapterPtrPtr
        ch: cython.pointer[lib.AVChapter]
        entry: dict

        with cython.nogil:
            _free_chapters(self.ptr)

        ch_array = cython.cast(
            AVChapterPtrPtr,
            lib.av_malloc(count * cython.sizeof(cython.pointer[lib.AVChapter])),
        )
        if ch_array == cython.NULL:
            raise MemoryError("av_malloc failed for chapters")

        for i in range(count):
            entry = chapters[i]
            ch = cython.cast(
                cython.pointer[lib.AVChapter],
                lib.av_malloc(cython.sizeof(lib.AVChapter)),
            )
            if ch == cython.NULL:
                raise MemoryError("av_malloc failed for chapter")
            ch.id = entry["id"]
            ch.start = cython.cast(int64_t, entry["start"])
            ch.end = cython.cast(int64_t, entry["end"])
            to_avrational(entry["time_base"], cython.address(ch.time_base))
            ch.metadata = cython.NULL
            if "metadata" in entry:
                dict_to_avdict(
                    cython.address(ch.metadata),
                    entry["metadata"],
                    self.metadata_encoding,
                    self.metadata_errors,
                )
            ch_array[i] = ch

        self.ptr.nb_chapters = count
        self.ptr.chapters = ch_array


def open(
    file,
    mode=None,
    format=None,
    options=None,
    container_options=None,
    stream_options=None,
    metadata_encoding="utf-8",
    metadata_errors="strict",
    buffer_size=32768,
    timeout=None,
    io_open=None,
    hwaccel=None,
):
    """open(file, mode='r', **kwargs)

    Main entrypoint to opening files/streams.

    :param str file: The file to open, which can be either a string or a file-like object.
    :param str mode: ``"r"`` for reading and ``"w"`` for writing.
    :param str format: Specific format to use. Defaults to autodect.
    :param dict options: Options to pass to the container and all streams.
    :param dict container_options: Options to pass to the container.
    :param list stream_options: Options to pass to each stream.
    :param str metadata_encoding: Encoding to use when reading or writing file metadata.
        Defaults to ``"utf-8"``.
    :param str metadata_errors: Specifies how to handle encoding errors; behaves like
        ``str.encode`` parameter. Defaults to ``"strict"``.
    :param int buffer_size: Size of buffer for Python input/output operations in bytes.
        Honored only when ``file`` is a file-like object. Defaults to 32768 (32k).
    :param timeout: How many seconds to wait for data before giving up, as a float, or a
        ``(open timeout, read timeout)`` tuple.
    :param callable io_open: Custom I/O callable for opening files/streams.
        This option is intended for formats that need to open additional
        file-like objects to ``file`` using custom I/O.
        The callable signature is ``io_open(url: str, flags: int, options: dict)``, where
        ``url`` is the url to open, ``flags`` is a combination of AVIO_FLAG_* and
        ``options`` is a dictionary of additional options. The callable should return a
        file-like object.
    :param HWAccel hwaccel: Optional settings for hardware-accelerated decoding.
    :rtype: Container

    For devices (via ``libavdevice``), pass the name of the device to ``format``,
    e.g.::

        >>> # Open webcam on MacOS.
        >>> av.open('0', format='avfoundation') # doctest: +SKIP

    For DASH and custom I/O using ``io_open``, add a protocol prefix to the ``file`` to
    prevent the DASH encoder defaulting to the file protocol and using temporary files.
    The custom I/O callable can be used to remove the protocol prefix to reveal the actual
    name for creating the file-like object. E.g.::

        >>> av.open("customprotocol://manifest.mpd", "w", io_open=custom_io) # doctest: +SKIP

    .. seealso:: :ref:`garbage_collection`

    More information on using input and output devices is available on the
    `FFmpeg website <https://www.ffmpeg.org/ffmpeg-devices.html>`_.
    """

    if not (mode is None or (isinstance(mode, str) and mode == "r" or mode == "w")):
        raise ValueError(f"mode must be 'r', 'w', or None, got: {mode}")

    if isinstance(file, str):
        pass
    elif isinstance(file, Path):
        file = f"{file}"
    elif mode is None:
        mode = getattr(file, "mode", None)

    if mode is None:
        mode = "r"

    if isinstance(timeout, tuple):
        if not len(timeout) == 2:
            raise ValueError("timeout must be `float` or `tuple[float, float]`")

        open_timeout, read_timeout = timeout
    else:
        open_timeout = timeout
        read_timeout = timeout

    if mode.startswith("r"):
        return InputContainer(
            _cinit_sentinel,
            file,
            format,
            options,
            container_options,
            stream_options,
            hwaccel,
            metadata_encoding,
            metadata_errors,
            buffer_size,
            open_timeout,
            read_timeout,
            io_open,
        )

    if stream_options:
        raise ValueError(
            "Provide stream options via Container.add_stream(..., options={})."
        )
    return OutputContainer(
        _cinit_sentinel,
        file,
        format,
        options,
        container_options,
        stream_options,
        None,
        metadata_encoding,
        metadata_errors,
        buffer_size,
        open_timeout,
        read_timeout,
        io_open,
    )



from enum import Flag

import cython
import cython.cimports.libav as lib
from cython.cimports.av.descriptor import wrap_avclass

_cinit_bypass_sentinel = cython.declare(object, object())


@cython.cfunc
def build_container_format(
    iptr: cython.pointer[cython.const[lib.AVInputFormat]],
    optr: cython.pointer[cython.const[lib.AVOutputFormat]],
) -> ContainerFormat:
    if not iptr and not optr:
        raise ValueError("needs input format or output format")
    format: ContainerFormat = ContainerFormat.__new__(
        ContainerFormat, _cinit_bypass_sentinel
    )
    format.iptr = iptr
    format.optr = optr
    format.name = optr.name if optr else iptr.name
    return format


# fmt: off
class Flags(Flag):
    no_file = lib.AVFMT_NOFILE
    need_number: "Needs '%d' in filename." = lib.AVFMT_NEEDNUMBER
    show_ids: "Show format stream IDs numbers." = lib.AVFMT_SHOW_IDS
    global_header: "Format wants global header." = lib.AVFMT_GLOBALHEADER
    no_timestamps: "Format does not need / have any timestamps." = lib.AVFMT_NOTIMESTAMPS
    generic_index: "Use generic index building code." = lib.AVFMT_GENERIC_INDEX
    ts_discont: "Format allows timestamp discontinuities" = lib.AVFMT_TS_DISCONT
    variable_fps: "Format allows variable fps." = lib.AVFMT_VARIABLE_FPS
    no_dimensions: "Format does not need width/height" = lib.AVFMT_NODIMENSIONS
    no_streams: "Format does not require any streams" = lib.AVFMT_NOSTREAMS
    no_bin_search: "Format does not allow to fall back on binary search via read_timestamp" = lib.AVFMT_NOBINSEARCH
    no_gen_search: "Format does not allow to fall back on generic search" = lib.AVFMT_NOGENSEARCH
    no_byte_seek: "Format does not allow seeking by bytes" = lib.AVFMT_NO_BYTE_SEEK
    ts_nonstrict: "Format does not require strictly increasing timestamps, but they must still be monotonic." = lib.AVFMT_TS_NONSTRICT
    ts_negative: "Format allows muxing negative timestamps." = lib.AVFMT_TS_NEGATIVE
    # If not set the timestamp will be shifted in `av_write_frame()` and `av_interleaved_write_frame()`
    # so they start from 0. The user or muxer can override this through AVFormatContext.avoid_negative_ts
    seek_to_pts: "Seeking is based on PTS" = lib.AVFMT_SEEK_TO_PTS
# fmt: on


@cython.cclass
class ContainerFormat:
    """Descriptor of a container format.

    :param str name: The name of the format.
    :param str mode: ``'r'`` or ``'w'`` for input and output formats; defaults
        to None which will grab either.

    """

    def __cinit__(self, name, mode=None):
        if name is _cinit_bypass_sentinel:
            return

        # We need to hold onto the original name because AVInputFormat.name is
        # actually comma-separated, and so we need to remember which one this was.
        self.name = name

        # Searches comma-separated names.
        if mode is None or mode == "r":
            self.iptr = lib.av_find_input_format(name)

        if mode is None or mode == "w":
            self.optr = lib.av_guess_format(name, cython.NULL, cython.NULL)

        if not self.iptr and not self.optr:
            raise ValueError(f"no container format {name!r}")

    def __repr__(self):
        return f"<av.{self.__class__.__name__} {self.name!r}>"

    @property
    def descriptor(self):
        if self.iptr:
            return wrap_avclass(self.iptr.priv_class)
        else:
            return wrap_avclass(self.optr.priv_class)

    @property
    def options(self):
        return self.descriptor.options

    @property
    def input(self):
        """An input-only view of this format."""
        if self.iptr == cython.NULL:
            return None
        elif self.optr == cython.NULL:
            return self
        else:
            return build_container_format(self.iptr, cython.NULL)

    @property
    def output(self):
        """An output-only view of this format."""
        if self.optr == cython.NULL:
            return None
        elif self.iptr == cython.NULL:
            return self
        else:
            return build_container_format(cython.NULL, self.optr)

    @property
    def is_input(self):
        return self.iptr != cython.NULL

    @property
    def is_output(self):
        return self.optr != cython.NULL

    @property
    def long_name(self):
        # We prefer the output names since the inputs may represent
        # multiple formats.
        return self.optr.long_name if self.optr else self.iptr.long_name

    @property
    def extensions(self):
        exts: set = set()
        if self.iptr and self.iptr.extensions:
            exts.update(self.iptr.extensions.split(","))
        if self.optr and self.optr.extensions:
            exts.update(self.optr.extensions.split(","))
        return exts

    @property
    def flags(self):
        """
        Get the flags bitmask for the format.

        :rtype: int
        """
        return (self.iptr.flags if self.iptr else 0) | (
            self.optr.flags if self.optr else 0
        )

    @property
    def no_file(self):
        return bool(self.flags & lib.AVFMT_NOFILE)


@cython.cfunc
def get_output_format_names() -> set:
    names: set = set()
    ptr: cython.pointer[cython.const[lib.AVOutputFormat]]
    opaque: cython.p_void = cython.NULL
    while True:
        ptr = lib.av_muxer_iterate(cython.address(opaque))
        if ptr:
            names.add(ptr.name)
        else:
            break
    return names


@cython.cfunc
def get_input_format_names() -> set:
    names: set = set()
    ptr: cython.pointer[cython.const[lib.AVInputFormat]]
    opaque: cython.p_void = cython.NULL
    while True:
        ptr = lib.av_demuxer_iterate(cython.address(opaque))
        if ptr:
            names.add(ptr.name)
        else:
            break
    return names


formats_available = get_output_format_names()
formats_available.update(get_input_format_names())
format_descriptor = wrap_avclass(lib.avformat_get_class())
